-------------------------------------------------------------------------------
-- Title      : RS232 UART template
-- Project    : Praktikum "Serielle Kommunikation"
-------------------------------------------------------------------------------
-- File       : receive.vhd
-- Author     : DI. R. Hecht  <ronald.hecht@etechnik.uni-rostock.de>
-- Company    : 
-- Created    : 2003-06-02
-- Last update: 2004-06-16
-- Platform   : 
-- Standard   : VHDL'87
-------------------------------------------------------------------------------
-- Description: 
-------------------------------------------------------------------------------
-- Copyright (c) 2003 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2003-06-02  1.0      hr55    Created
-- 2010-01-21               at101   Ported to Digilent Nexys 2 Board, added process to control 7 segment display
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use work.config.all;
use work.makros.all;
use work.crc_pack.all;

entity receive is

  generic (
    BAUD_RATE : integer := 38400);
  port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      an        : out std_logic_vector(3 downto 0);  -- DO NOT CHANGE THIS: an bits are needed for refreshing seven-segment displays
	  seg_data  : out std_logic_vector(6 downto 0);  -- DO NOT CHANGE THIS: seg ist used to send signs to the seven-segment displays
      noise     : in  std_logic_vector(7 downto 0);
      err       : out std_logic;
      rdy       : out std_logic;
	  rx        : in  std_logic;
      seg_lo_tb : out std_logic_vector(6 downto 0);
      seg_hi_tb : out std_logic_vector(6 downto 0)
  );
end receive;

architecture rtl of receive is

  -- bitrate for transmit
  constant TX_TICKS_PER_SECOND : integer := div_round(CLK_RATE, BAUD_RATE);
  -- oversampling ratio of 8 for receive
  constant RX_TICKS_PER_SECOND : integer := div_round(CLK_RATE, BAUD_RATE * 8);
  -- Clock ticks for driving seven-segment displays
  constant SEG_TICKS_PER_SECOND : integer := 100;  -- 100 Hz refresh scheme: Refresh every 10 ms

  -- input for the seven-segment display controller
  signal seg_lo_sig : std_logic_vector(6 downto 0);
  signal seg_hi_sig : std_logic_vector(6 downto 0);
  
begin  -- rtl
  seg_lo_tb <= seg_lo_sig;
  seg_hi_tb <= seg_hi_sig;
	
	
  -- insert your code here
	

  ------------------------------------------------------------------------------------------------------------------------------------------
  -- DO NOT CHANGE THIS 
  -- purpose: Refresh mechanism for seven-segment-display
  refresh : process (clk, rst)
    -- seven-segment display clk scaler and tick pulses
    variable seg_tick   : std_logic_vector(3 downto 0);
    variable seg_scaler : integer range 0 to CLK_RATE / SEG_TICKS_PER_SECOND - 1;
  begin  -- process refresh
	if rising_edge(clk) then            -- rising clock edge
      -- scale clk 
      if seg_scaler /= CLK_RATE / SEG_TICKS_PER_SECOND - 1 then
        -- Assert seg_tick0 for activating first seven-segment display (low active)
        if seg_scaler < (CLK_RATE / SEG_TICKS_PER_SECOND)/4 then
          seg_tick(0) := '0';
          seg_tick(1) := '1';
          seg_tick(2) := '1';
          seg_tick(3) := '1';
        -- Assert seg_tick1 for activating first seven-segment display (low active)  
        elsif seg_scaler < 2*(CLK_RATE / SEG_TICKS_PER_SECOND)/4 then
          seg_tick(0) := '1';
          seg_tick(1) := '0';
          seg_tick(2) := '1';
          seg_tick(3) := '1';
        -- Assert seg_tick2 for activating first seven-segment display (low active)  
        elsif seg_scaler < 3*(CLK_RATE / SEG_TICKS_PER_SECOND)/4 then
          seg_tick(0) := '1';
          seg_tick(1) := '1';
          seg_tick(2) := '0';
          seg_tick(3) := '1';
        -- Assert seg_tick3 for activating first seven-segment display (low active)  
        else
          seg_tick(0) := '1';
          seg_tick(1) := '1';
          seg_tick(2) := '1';
          seg_tick(3) := '0';
        end if;
        seg_scaler := seg_scaler + 1;
      else
        seg_scaler := 0;
        -- Assert seg_tick0 for a last time if desired
        seg_tick(0)  := '0';
        seg_tick(1)  := '1';
        seg_tick(2)  := '1';
        seg_tick(3)  := '1';
      end if;  
      if sync_reset(rst) then
        seg_scaler := 0;
        seg_tick(0)  := '1';
        seg_tick(1)  := '1';
        seg_tick(2)  := '1';
        seg_tick(3)  := '1';
      end if;
    end if;		-- clk
    if async_reset(rst) then
      seg_scaler := 0;
      seg_tick(0)  := '1';
      seg_tick(1)  := '1';
      seg_tick(2)  := '1';
      seg_tick(3)  := '1';
    end if;
	
	-- seven-segment display is low active
	seg_data <= (others => '1');
	if seg_tick(0) = '0' then
	  seg_data <= not seg_lo_sig;
	end if;
	if seg_tick(1) = '0' then
	  seg_data <= not seg_hi_sig;
	end if;
	
    --  DO NOT CHANGE THIS: Drive seven-segment displays
    an(0) <= seg_tick(0);
    an(1) <= seg_tick(1);
    an(2) <= seg_tick(2);
    an(3) <= seg_tick(3);
	
  end process refresh;
  
end rtl;
