-------------------------------------------------------------------------------
-- Title      : Testbench for design "transmit" and "receive"
-- Project    : 
-------------------------------------------------------------------------------
-- File       : sercom_tb.vhd
-- Author     : DI. R. Hecht  <ronald.hecht@etechnik.uni-rostock.de>
-- Company    : 
-- Created    : 2003-07-04
-- Last update: 2004-08-12
-- Platform   : 
-- Standard   : VHDL'87
-------------------------------------------------------------------------------
-- Description: 
-------------------------------------------------------------------------------
-- Copyright (c) 2003 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2003-07-04  1.0      hr55	Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use work.makros.all;
use work.crc_pack.all;
use work.syslog.all;

entity sercom_tb is

  generic (
    period    : time := 1 sec / CLK_RATE;
    hold_time : time := 1 ns;
    
    -- uart baud rate
    BAUD_RATE : integer := 38400);
 
end sercom_tb;

architecture beh of sercom_tb is

  -- time for one uart bit
  constant BIT_PERIOD : time := 1 sec / BAUD_RATE;

  component transmit
    generic (
      BAUD_RATE : integer);
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      an        : out std_logic_vector(3 downto 0);  -- DO NOT CHANGE THIS: an bits are needed for refreshing seven-segment displays
	  seg_data  : out std_logic_vector(6 downto 0);  -- DO NOT CHANGE THIS: seg_data ist used to send signs to the seven-segment displays
      a         : in  std_logic_vector(3 downto 0);
      b         : in  std_logic_vector(3 downto 0);
      send      : in  std_logic;            -- show product and transmit
      ack       : out std_logic;			   -- testbench synchronisation
      tx        : out std_logic;           -- RS232 transmit
      seg_lo_tb : out std_logic_vector(6 downto 0);  -- product input for testbench
      seg_hi_tb : out std_logic_vector(6 downto 0)   -- product input for testbench
	);
  end component;

  component receive
    generic (
      BAUD_RATE : integer);
    port (
      clk       : in  std_logic;
      rst       : in  std_logic;
      an        : out std_logic_vector(3 downto 0);  -- DO NOT CHANGE THIS: an bits are needed for refreshing seven-segment displays
	  seg_data  : out std_logic_vector(6 downto 0);  -- DO NOT CHANGE THIS: seg ist used to send signs to the seven-segment displays
      noise     : in  std_logic_vector(7 downto 0);
      err       : out std_logic;
      rdy       : out std_logic;
	  rx        : in  std_logic;
      seg_lo_tb : out std_logic_vector(6 downto 0);
      seg_hi_tb : out std_logic_vector(6 downto 0)
    );
  end component;
  
  -- component ports
  signal clk    : std_logic := '0';
  signal rst    : std_logic := '1';
  signal a      : std_logic_vector(3 downto 0) := (others => '0');
  signal b      : std_logic_vector(3 downto 0) := (others => '0');
  signal send   : std_logic;
  signal ack    : std_logic;
  signal seg_lo : std_logic_vector(6 downto 0);
  signal seg_hi : std_logic_vector(6 downto 0);
  signal tx     : std_logic;

  signal noise  : std_logic_vector(7 downto 0) := "00000000";
  signal err    : std_logic;
  signal rdy    : std_logic;
  signal rx     : std_logic;
  signal rx_seg_lo : std_logic_vector(6 downto 0);
  signal rx_seg_hi : std_logic_vector(6 downto 0);
  
  signal crc_s : std_logic_vector(4 downto 0);
  signal data_s : std_logic_vector(12 downto 0);
  
begin  -- beh

  -- component instantiation
  transmit_1: transmit
    generic map (
      BAUD_RATE => BAUD_RATE)
    port map (
      clk       => clk,
      rst       => rst,
      an        => open,
	  seg_data  => open,
      a         => a,
      b         => b,
      send      => send,
      ack       => ack,
      tx        => tx,
      seg_lo_tb => seg_lo,
      seg_hi_tb => seg_hi
	);

  receive_1: receive
    generic map (
      BAUD_RATE => BAUD_RATE)
    port map (
      clk       => clk,
      rst       => rst,
	  an        => open,
	  seg_data  => open,
      noise     => noise,
      err       => err,
      rdy       => rdy,
	  rx        => rx,
      seg_lo_tb => rx_seg_lo,
      seg_hi_tb => rx_seg_hi
    );

  rx <= tx;
  
  -- clock and reset generation
  clk <= not clk  after period / 2;
  rst <= '1', '0' after period * 4;

  stimuli: process

    procedure mul_and_send (n, m : in integer range 0 to 15) is
    begin 
      a <= to_unsigned_std_logic_vector(n, a'length) after hold_time;
      b <= to_unsigned_std_logic_vector(m, b'length) after hold_time;
      wait on clk until clk = '1';
      send <= '1' after hold_time;
      wait on clk until clk = '1' and ack = '1';
      send <= '0' after hold_time;
    end mul_and_send;
    
  begin  -- process tester
    
    -- do nothing during reset
    wait until rst = '0';

    for i in 0 to 15 loop
      for j in 0 to 15 loop
        mul_and_send(i, j);
      end loop;  -- j
    end loop;  -- i

    -- dummy send
    mul_and_send(0, 0);
    mul_and_send(0, 0);
    -- finish
    syslog_terminate;
    
  end process stimuli;


  tester: process

    procedure receive (n : out integer range 0 to 255) is
    begin 
      wait on clk until clk = '1' and rdy = '1';
      n := (seg2int(rx_seg_hi)*16 + seg2int(rx_seg_lo));
      if err = '1' then
        syslog(error, "Receive error");
      end if;
    end receive;

    variable result : integer range 0 to 255;
    
  begin  -- process tester
    
    for i in 0 to 15 loop
      for j in 0 to 15 loop
        receive(result);
        if result /= i * j then
          syslog(error, "Received bad product "
                 & image(i) & " * " & image(j)
                 & " /= " & image(result));
        else
          syslog(debug, "Received product "
                 & image(i) & " * " & image(j)
                 & " = " & image(result));
          
        end if;
      end loop;  -- j
    end loop;  -- i

    wait;
    
  end process tester;

  -- receive all characters until line-feed
  receiver : process

    -- serial data
    variable data : std_logic_vector(12 downto 0);
    variable crc : std_logic_vector(4 downto 0);
    
  begin  -- process receiver
 
      -- endless loop, can be restarted with "exit reset_handler"
    reset_handler : loop
      -- wait for start bit
      wait on tx until tx = '0';
      -- sample at middle
      wait for BIT_PERIOD / 2;
      -- and verify = '0'
      if tx /= '0' then
        syslog(warning, "Bad start bit");
        exit reset_handler;
      end if;
      -- sample all data bits
      for i in 12 downto 0 loop
        wait for BIT_PERIOD;
        data(i) := tx;
      end loop;  -- i
      -- sample stop bit
      wait for BIT_PERIOD;
      -- and verify = '1'
      if tx /= '1' then
        syslog(warning, "Bad stop bit");
        exit reset_handler;
      end if;
      crc := (others => '0');
      crc := next_crc(crc, data, CRC5_GEN);
      crc_s <= crc;
      data_s <= data;
      if crc /= "00000" then
        syslog(error, "Bad CRC" & std_logic'image(data(0)));
      end if;
    end loop;
    
  end process receiver;
 
end beh;

configuration sercom_tb_beh_cfg of sercom_tb is
  for beh
  end for;
end sercom_tb_beh_cfg;
